"""Utilities for reading and writing lexicon and/or
   segmentations. Probably you need to use the "model" module to
   create a model after reading in the segmentation files.

    Author: Javad Nouri 
    first.last  <at> cs.helsinki.fi
"""
from operator import itemgetter
import re
import random
import numpy as np
from tabulate import tabulate
from .util import sort_after_first

encoding = 'utf-8'
comment_pattern = re.compile(ur'[;#].*$', re.MULTILINE)
TABLE_FMT = 'orgtbl'
FLOAT_FMT = '.2f'
verbose = False



def read_simple_segmentation_file(file_name):
    with open(file_name, 'r') as fl:
        lines = fl.readlines()
        lines = [l.decode(encoding).strip().split('+-/ ') for l in lines]
        return lines

def process_seg(seg):
    ""
    return [seg[i] if i % 2 == 0 else int(seg[i]) for i in range(len(seg))]

def read_segmentation_file(segFileName):
    segmentationsP = [l.decode('utf-8').strip().split('\t')[0].strip().split() for l in open(segFileName, 'r').readlines()]
    segmentations = [process_seg(l) for l in segmentationsP]
    return segmentations

def read_lex_file(lexFileName, old=False):
    "Reads in a lexicon file and returns a lexicon object as well as number of classes used"
    splits = re.split('(Class [0-9]+)', open(lexFileName, 'r').read().decode('utf-8'))
    classPattern = re.compile(u'^Class [0-9]+$')
    cls = [[splits[i], splits[i+1]] for i in range(1, len(splits), 2)] 
    for c in cls:
        assert re.match(classPattern, c[0])
    cls = sorted([[int(c[0].replace('Class ', '')), [x.split() for x in c[1].split('\n') if len(x)]] for c in cls])
    
    lex = [{morphs[0]:float(morphs[1]) for morphs in c[1]} for c in cls]
    lex_comp = None
    try:
        lex_comp = [{morphs[0]:float(morphs[2]) for morphs in c[1]} for c in cls]
    except:
        lex_comp = None
    if old:
        lex = [{}] + lex + [{}]
        lex_comp = [{}] + lex_comp + [{}]
    return [lex, len(lex) - 2, lex_comp]

    
def print_lex(lex):
    "Print lexicon in a user friendly way"
    classCounter = 0
    for cl in lex:
        print 'Class', classCounter
        li = sorted(sorted([[key, value] for (key, value) in cl.items()], key=itemgetter(0)), key=itemgetter(1), reverse=True)
        print '\n'.join(['%s %.1f'% (l[0].encode('utf-8'),l[1]) for l in li])
        if len(cl) > 0:
            print 
        classCounter += 1

def print_transition_org(transitions):
    print '\n'.join(['|' + '|'.join([str(x) for x in l]) + '|' for l in transitions])

def print_transitions(transitions):
    return np.array(transitions)
    #return  '\n'.join(['|' + '|'.join([str(x) for x in l]) + '|' for l in transitions])

def print_segmentations(segs):
    return '\n'.join([print_word_segmentation(seg) for seg in segs])
        
def print_word_segmentation(seg):
    return u' '.join([unicode(x) for x in seg])


def utabulate(lst, headers=None, tablefmt=TABLE_FMT):
    """Create a tabulated version of the matrix "lst" in the specified
    format (default format specified by TABLE_FMT) with header (if
    provided)

    """
    if headers:
        return tabulate(unicodize(lst), headers=headers, tablefmt=tablefmt, floatfmt=FLOAT_FMT)
    else:
        return tabulate(unicodize(lst), tablefmt=tablefmt, floatfmt=FLOAT_FMT)




def print_error(err_text):
    """
    Utility method for printing error messages. 
    """
    print >>sys.stderr, err_text

def print_if_verbose(text):
    """Used for verbose mode.  Todo: Switch to python "Logging" next.

    """
    if verbose:
        print unicode(text).encode(encoding)


def unicodize(matrix):
    """Create a unicode representation of the elements of matrix

    """
    return [[unicode(el) for el in row] for row in matrix]


encoding    = "utf-8"
    
def read_label_file(labels_file):
    """Read label file and allowed choices.  A label file is a list (lisp
    style) of the form

    (
     (A 4 0 1)...
    )

    where A is the label, "4" shows that it represents is a 4-way
    dilemma, and 0,1 represent allowed choices for the dilemma.  Refer
    to the paper for details.

    """
    lines = []
    norm_pattern = re.compile(ur'\s+', re.MULTILINE)
    label_format = re.compile(ur'\(("."(?: [0-9])+)\)+', re.MULTILINE)

    pure_lisp = None
    with open(labels_file, 'r') as l_file:
        pure_lisp = re.sub(comment_pattern, '', l_file.read().decode(encoding))
        pure_lisp = re.sub(norm_pattern, ' ', pure_lisp)
        #print_if_verbose(pure_lisp, verbose)
    labels = {l.split()[0].replace("\"", ""):sort_after_first([int(x) for x in l.split()[1:]]) for l in re.findall(label_format, pure_lisp)}
    return labels


def read_gold_data(gold_file, labels_file):
    """Read the gold standard data which is annotated based on label
    file. 

    """
    labels = read_label_file(labels_file) #The labels and allowed choices 
    with open(gold_file, 'r') as g_file:
        lines = [l.strip().decode(encoding) for l in g_file.readlines()] 
        lines = [re.sub(comment_pattern, '', l) for l in lines]           # Remove comment content
        lines = [re.sub('[!,].+', '', l).strip() for l in lines]          # Remove everything after "!". "!" denotes a problem in annotation
                                                                          # which needs to be fixed. During the evaluation they are ignored.
                                                                          
        lines = [re.sub('[/-]', '+', l) for l in lines]                   # Replace '/' and '-' occurences with '+'.
                                                                          # / and - are used to annotate word compound boundaries and
                                                                          # derivational boundaries. They will be considered in the future,
                                                                          # for now we consider all types of boundaries the same: marked
                                                                          # with a "+".
                                                                          
        lines = filter(lambda x:len(x)>0, lines)                          # filter out empty lines.
        
     
    print_if_verbose('regex: ' + '^[' + ''.join(labels.keys()) + ' 	]+$')
            
    label_regex = re.compile('^[' + ''.join(labels.keys()) + ' 	]+$', re.MULTILINE)    #Regex to recognize label lines
    word_regex = re.compile(ur'^[0-9]+[ 	]+[^ .	]+$', re.MULTILINE)            #Regex to recognize segmentation lines
    prev_line = None   #Will hold the previous line
    gold_lines = []
    for line in lines:
        if '.' in line: #If there is a . in the line, it is an ambiguous segmentation line. 
            #print prev_line,":", line
            assert prev_line is not None #There must be a label line immediately before this line.
            if verbose:
                print_if_verbose(line + ' is ambigious with label ' + prev_line)
            gold_lines.append((prev_line, line,))
            prev_line = None
    	elif re.match(label_regex, line):
            print_if_verbose(line + ' is label')
            prev_line = line
    	elif re.match(word_regex, line):
            print_if_verbose(line +' is non-ambiguous')
            gold_lines.append((None,line,))
            prev_line = None
        else:
            print_error('Error |' + line + '| is unknown. Check for any labels not explained')
            raise RuntimeError()
    return gold_lines, labels


def read_and_validate_gold(gold_file, label_file):
    """
    """
    gold_lines, labels = read_gold_data(gold_file, label_file) #Read from file
    #Clean the segmentation lines:
    gold_lines = [(label, re.sub('^[0-9 	]+', '', seg))for label, seg in gold_lines]
    
    #split labels and segmentations
    gold_lines = [(label.split() if label else [], seg.split('.'))for label, seg in gold_lines]

    #Check length of labels
    for label, seg in gold_lines:
        validate_seg_and_label(seg, label, labels)
    #seg_line = re.sub('^[0-9 	]+', '', seg_line)
    return gold_lines, labels
    


def validate_seg_and_label(seg, label, labels):
    """Validate segmentation and label. Make sure that all the labels used
    in the annotated segmentation are valid labels that appear in the
    labels file, and that they are used properly: e.g. for each
    instance of a 4-way dilemma, the label must have been used 2
    times, etc.

    """
    assert len(label) == len(seg)-1
    if label:
        for l in label:
            assert l in labels
            
